import itertools
import os
from collections import defaultdict
from multiprocessing import cpu_count
import networkx as nx
import matplotlib.pyplot as plt
import numpy as np
from multiprocessing import Pool
from projectq.cengines import BasicEngine
from projectq.backends import ResourceCounter
from projectq import MainEngine
from projectq.ops import QubitOperator, All, H, Rx, Measure, MatrixGate, Rz, Rzz, Z
from projectq.setups import linear

from Qcover.backends import Backend
from Qcover.utils import get_graph_weights


class CircuitByProjectq(Backend):
    """generate a instance of CircuitByProjectq"""
    def __init__(self,
                 research: str = "QAOA",
                 is_parallel: bool = None) -> None:
        """initialize a instance of CircuitByProjectq"""
        super(CircuitByProjectq, self).__init__()

        self._p = None
        self._origin_graph = None
        self._is_parallel = False if is_parallel is None else is_parallel
        self._research = research

        self._nodes_weight = None
        self._edges_weight = None
        self._element_to_graph = None
        self._pargs = None
        self._expectation_path = []
        self._element_expectation = dict()

    @property
    def element_expectation(self):
        return self._element_expectation

    @staticmethod
    def get_operator(element):
        op = QubitOperator()
        if isinstance(element, int):
            op += QubitOperator('Z' + str(element))
        else:
            op += QubitOperator('Z' + str(element[0])) + QubitOperator('Z' + str(element[1]))
        return op

    def get_QAOA_circuit(self, p, graph, node_to_qubit):
        gamma_list, beta_list = self._pargs[: p], self._pargs[p:]
        eng = MainEngine()
        qubits = eng.allocate_qureg(len(graph.nodes))
        All(H) | qubits

        for k in range(p):
            for edge in graph.edges:
                u, v = node_to_qubit[edge[0]], node_to_qubit[edge[1]]
                if u == v:
                    continue
                Rzz(2 * gamma_list[k] * self._edges_weight[edge[0], edge[1]]) | (qubits[u], qubits[v])

            for nd in graph.nodes:
                u = node_to_qubit[nd]
                Rz(2 * gamma_list[k] * self._nodes_weight[nd]) | qubits[u]
                Rx(2 * beta_list[k]) | qubits[u]

        # print("before flush")
        eng.flush()
        # print("after flush")
        # print("the original element is", original_e)
        # assert len(qubits) == len(node_list)
        return eng, qubits

    def get_GHZ_circuit(self, p, graph, node_to_qubit):
        gamma_list, beta_list = self._pargs[: p], self._pargs[p:]
        pivot = len(self._origin_graph) // 2

        eng = MainEngine()
        qubits = eng.allocate_qureg(len(graph.nodes))
        All(H) | qubits

        for k in range(p):
            for edge in graph.edges:
                u, v = node_to_qubit[edge[0]], node_to_qubit[edge[1]]
                if u == v:
                    continue
                Rzz(2 * gamma_list[k] * self._edges_weight[edge[0], edge[1]]) | (qubits[u], qubits[v])

            for nd in graph.nodes:
                if nd != pivot:
                    u = node_to_qubit[nd]
                    Rx(2 * beta_list[k]) | qubits[u]

        # print("before flush")
        eng.flush()
        # print("after flush")
        # print("the original element is", original_e)
        # assert len(qubits) == len(node_list)
        return eng, qubits

    def get_expectation(self, element_graph, p=None):
        """
        transform the graph to circuit according to the computing_framework
        Args:
            graph (nx.Graph): graph to be transformed to circuit
            params (np.array): Optimal parameters
            original_e (Optional[None, int, tuple])
        Return:
            if original_e=None, then the graph is the whole original graph generated by
            generate_weighted_graph(), so just return the circuit transformed by it

            if original_e is a int, then the subgraph is generated by node(idx = original_e
            in whole graph), so return the it's idx mapped by node_to_qubit[], and the circuit

            if original_e is a tuple, then the subgraph is generated by edge(node idx = original_e
            in whole graph), so return the it's idx mapped by node_to_qubit[] as
            tuple(mapped node_id1, mapped node_id2), and the circuit
        """
        if self._is_parallel is False:
            p = self._p if p is None else p
            original_e, graph = element_graph
        else:
            p = self._p if len(element_graph) == 1 else element_graph[1]
            original_e, graph = element_graph[0]

        node_to_qubit = defaultdict(int)
        node_list = list(graph.nodes)
        for i in range(len(node_list)):
            node_to_qubit[node_list[i]] = i

        if self._research == "QAOA":
            eng, qubits = self.get_QAOA_circuit(p, graph, node_to_qubit)
        elif self._research == "GHZ":
            pivot = len(self._origin_graph) // 2
            if pivot in graph:
                eng, qubits = self.get_GHZ_circuit(p, graph, node_to_qubit)
            else:
                eng, qubits = self.get_QAOA_circuit(p, graph, node_to_qubit)

        if isinstance(original_e, int):
            weight = self._nodes_weight[original_e]
            op = self.get_operator(node_to_qubit[original_e])
        else:
            weight = self._edges_weight[original_e]
            op = self.get_operator((node_to_qubit[original_e[0]], node_to_qubit[original_e[1]]))

        exp_res = eng.backend.get_expectation_value(op, qubits)
        All(Measure) | qubits
        eng.flush(deallocate_qubits=True)        #

        return weight, exp_res

    def expectation_calculation(self, p=None):
        if self._nodes_weight is None or self._edges_weight is None:
            nodes_weight, edges_weight = get_graph_weights(self._origin_graph)
            self._nodes_weight, self._edges_weight = nodes_weight, edges_weight

        self._element_expectation = {}
        if self._is_parallel:
            return self.expectation_calculation_parallel(p)
        else:
            return self.expectation_calculation_serial(p)

    def expectation_calculation_serial(self, p=None):
        cpu_num = cpu_count()  # 自动获取最大核心数目
        os.environ['OMP_NUM_THREADS'] = str(cpu_num)
        os.environ['OPENBLAS_NUM_THREADS'] = str(cpu_num)
        os.environ['MKL_NUM_THREADS'] = str(cpu_num)
        os.environ['VECLIB_MAXIMUM_THREADS'] = str(cpu_num)
        os.environ['NUMEXPR_NUM_THREADS'] = str(cpu_num)

        res = 0
        for item in self._element_to_graph.items():
            w_i, exp_i = self.get_expectation(item, p)
            if isinstance(item[0], tuple):
                self._element_expectation[item[0]] = exp_i
            res += w_i * exp_i

        print("Total expectation of original graph is: ", res)
        self._expectation_path.append(res)
        return res

    def expectation_calculation_parallel(self, p=None):
        cpu_num = 1
        os.environ['OMP_NUM_THREADS'] = str(cpu_num)
        os.environ['OPENBLAS_NUM_THREADS'] = str(cpu_num)
        os.environ['MKL_NUM_THREADS'] = str(cpu_num)
        os.environ['VECLIB_MAXIMUM_THREADS'] = str(cpu_num)
        os.environ['NUMEXPR_NUM_THREADS'] = str(cpu_num)

        circ_res = []
        args = list(itertools.product(self._element_to_graph.items(), [p]))
        pool = Pool(os.cpu_count())
        circ_res.append(pool.map(self.get_expectation, args))

        pool.terminate()  # pool.close()
        pool.join()
        res = 0
        for it in circ_res[0]:
            res += it[0] * it[1]

        # res = sum(circ_res[0])
        print("Total expectation of original graph is: ", res)
        self._expectation_path.append(res)
        return res

    def visualization(self):
        plt.figure()
        plt.plot(range(1, len(self._expectation_path) + 1), self._expectation_path, "ob-", label="projectq")
        plt.ylabel('Expectation value')
        plt.xlabel('Number of iterations')
        plt.legend()
        plt.show()



# 
# This is a quantum computer simulator.
# This is a special variant of Qton.
# This is for powering Qcover.
# 
# Author(s): Yunheng Ma
# Timestamp: 2022-03-30
# 


import numpy as np
from random import choices

# alphabet
alp = [
    'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
    'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B', 'C', 'D',
    'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S',
    'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
      ]

class Qcircuit:

    # num_qubits = 0
    # state = None
    def __init__(self, num_qubits=1, mode="statevector"):
        self.num_qubits = num_qubits
        self.state = np.zeros(2**num_qubits, complex)
        self.state[0] = 1.0
        self.mode = mode

    def apply_tensor(self, gate, num_qubits, *qubits):
        # based on `numpy.tensordot` and `numpy.einsum`
        if num_qubits > len(set(qubits)):
            raise Exception('Duplicate qubits in input.')
        global alp

        a_idx = [*range(num_qubits, 2 * num_qubits)]
        b_idx = [self.num_qubits - i - 1 for i in qubits]

        rep = gate.reshape([2] * 2 * num_qubits)
        self.state = self.state.reshape([2] * self.num_qubits)
        self.state = np.tensordot(rep, self.state, axes=(a_idx, b_idx))

        s = ''.join(alp[:self.num_qubits])
        end = s
        start = ''
        for i in range(num_qubits):
            start += end[self.num_qubits - qubits[i] - 1]
            s = s.replace(start[i], '')
        start = start + s
        self.state = np.einsum(start + '->' + end, self.state).reshape(-1)
        return None

    def _apply_1q_(self, gate, qubit):
        L = 2**qubit
        for i in range(0, 2**self.num_qubits, 2**(qubit + 1)):
            for j in range(i, i + L):
                self.state[j], \
                self.state[j + L] = np.matmul(gate, [
                    self.state[j],
                    self.state[j + L]
                    ])
        return None

    def _apply_2q_(self, gate, qubit1, qubit2):
        if qubit1 == qubit2:
            raise Exception('Cannot be same qubits.')
        if qubit1 > qubit2:
            q1, q2 = qubit1, qubit2
        else:
            q1, q2 = qubit2, qubit1
        L1, L2, L3 = 2**qubit2, 2**qubit1, 2**qubit2 + 2**qubit1
        for i in range(0, 2**self.num_qubits, 2**(q1 + 1)):
            for j in range(0, 2**q1, 2**(q2 + 1)):
                for k in range(0, 2**q2):
                    step = i + j + k
                    self.state[step], \
                    self.state[step + L1], \
                    self.state[step + L2], \
                    self.state[step + L3] = np.matmul(gate, [
                        self.state[step],
                        self.state[step + L1],
                        self.state[step + L2],
                        self.state[step + L3]
                        ])
        return None

    def apply(self, gate, num_qubits, *qubits, mode):
        if mode == "tensor":
            self.apply_tensor(gate, num_qubits, qubits)
        else:
            if num_qubits == 1:
                self._apply_1q_(gate, qubits[0])
            else:
                self._apply_2q_(gate, qubits[0], qubits[1])

    def z(self, qubit):
        gate = np.array([[1, 0], [0, -1.]])
        if self.mode == "tensor":
            self.apply_tensor(gate, 1, qubit)
        else:
            self._apply_1q_(gate, qubit)
        return gate

    def h(self, qubit):
        gate = np.array(
            [[1, 1.],
             [1, -1.]]) * np.sqrt(0.5)
        if self.mode == "tensor":
            self.apply_tensor(gate, 1, qubit)
        else:
            self._apply_1q_(gate, qubit)
        return gate

    def rz(self, qubit, theta):
        gate = np.array([
            [np.exp(-1j * theta * 0.5), 0],
            [0, np.exp(1j * theta * 0.5)],
        ])
        if self.mode == "tensor":
            self.apply_tensor(gate, 1, qubit)
        else:
            self._apply_1q_(gate, qubit)
        return gate

    def rx(self, qubit, theta):
        t = theta * 0.50
        gate = np.array([
            [np.cos(t), -1j * np.sin(t)], 
            [-1j * np.sin(t), np.cos(t)]])
        if self.mode == "tensor":
            self.apply_tensor(gate, 1, qubit)
        else:
            self._apply_1q_(gate, qubit)
        return gate

    def rzz(self, qubit1, qubit2, theta):
        a, b = np.exp(-0.5j*theta), np.exp(0.5j*theta)
        gate = np.diag([a, b, b, a])
        if self.mode == "tensor":
            self.apply_tensor(gate, 2, qubit1, qubit2)
        else:
            self._apply_2q_(gate, qubit1, qubit2)
        return gate

    def sample(self, shots=1024):
        p = self.state * self.state.conj()
        N = self.state.shape[0]
        memory = choices(range(N), weights=p, k=shots)
        counts = {}
        for i in memory:
            key = format(i, '0%db' % self.num_qubits)
            if key in counts:
                counts[key] += 1
            else:
                counts[key] = 1
        return counts



class Qcodes:
    # codes = []
    def __init__(self, num_qubits=1):
        self.codes = ['qc = Qcircuit(%d)'%num_qubits]
        return None


    def z(self, qubit):
        self.codes.append('qc.z(%d)'%qubit)
        return None 


    def h(self, qubit):
        self.codes.append('qc.h(%d)'%qubit)
        return None


    def rx(self, theta, qubit):
        self.codes.append('qc.rx(%f, %d)'%(theta, qubit))
        return None


    def rzz(self, theta, qubit1, qubit2):
        self.codes.append('qc.rzz(%f, %d, %d)'%(theta, qubit1, qubit2))
        return None


    def run(self):
        from Qcover.simulator import Qcircuit
        dic = locals()
        for code in self.codes:
            exec(code, dic)
        return dic['qc']



import sys
import numpy as np
from typing import Optional
from scipy import optimize as opt
from Qcover.optimizers import Optimizer
from Qcover.exceptions import ArrayShapeError, OptimizerConfigError


class Interp(Optimizer):
    """
        Interp optimizer: a heuristic optimization method for QAOA,
        implemented according to the paper
        "Quantum Approximate Optimization Algorithm: Performance, Mechanism, and Implementation on Near-Term Devices"
    """
    def __init__(self,
                 optimize_method='COBYLA',
                 options: dict = None, #{'maxiter':300, 'disp':True, 'rhobeg': 1.0, 'tol':1e-6},
                 initial_point: Optional[np.ndarray] = None):
        super().__init__()
        self._p = None
        self._optimize_method = optimize_method
        self._options = options
        self._initial_point = initial_point

    def _minimize(self, objective_function):
        """
        minimize the loss function
        Args:
            loss: the loss function
            initial_point: the init parameters of gamma and beta

        Returns:
            x: the optimized gamma and beta
            value: the optimized value of loss function
            nfev: is the number of objective function calls
        """
        nfev = 0
        for k in range(1, self._p + 1):
            if k == 1:
                # though only used in p=1, but to be consistent with other optimizers,
                # self._initial_point should be defined according to p
                gamma_list, beta_list = self._initial_point[: k], self._initial_point[self._p: self._p + k]
                gamma_list = np.insert(gamma_list, 0, 0)
                beta_list = np.insert(beta_list, 0, 0)
            else:
                gamma_list = np.insert(gamma_list, 0, 0)
                beta_list = np.insert(beta_list, 0, 0)
                gamma_list = np.append(gamma_list, 0)
                beta_list = np.append(beta_list, 0)
                gamma_list_new, beta_list_new = gamma_list, beta_list
                for i in range(1, k + 1):
                    gamma_list_new[i] = (i - 1) / k * gamma_list[i - 1] + (k - i + 1) / k * gamma_list[i]
                    beta_list_new[i] = (i - 1) / k * beta_list[i - 1] + (k - i + 1) / k * beta_list[i]

                    if gamma_list_new[i] < -np.pi / 2:
                        gamma_list_new[i] = -np.pi / 2 + 0.01
                    if beta_list_new[i] < -np.pi / 4:
                        beta_list_new[i] = -np.pi / 4 + 0.01
                    if gamma_list_new[i] > np.pi / 2:
                        gamma_list_new[i] = np.pi / 2 - 0.01
                    if beta_list_new[i] > np.pi / 4:
                        beta_list_new[i] = np.pi / 4 - 0.01

                gamma_list, beta_list = gamma_list_new, beta_list_new

            res = opt.minimize(objective_function,
                               x0=np.append(gamma_list[1:k+1], beta_list[1:k+1]),
                               args=k,
                               method=self._optimize_method,
                               jac=opt.rosen_der,
                               options=self._options)

            gamma_list, beta_list = res["x"][:k], res["x"][k:]
            value = res["fun"]
            nfev += res["nfev"]

            try:
                optv_f, aprv_f = "optimal_value" in self._options, "approximate_ratio" in self._options
                if optv_f ^ aprv_f:
                    raise OptimizerConfigError("optimal value and approximate ratio option should be set simultaneously to the optimizer")
                if optv_f & aprv_f:
                    optv, aprv = self._options["optimal_value"], self._options["approximate_ratio"]
                    if optv * aprv >= value:
                        print("The target approximation relative to the optimal value is reached. "
                              "The value of P is", k)
                        break
            except OptimizerConfigError as e:
                print(e)
                sys.exit()

        return np.append(gamma_list, beta_list), value, nfev
        # return {"gamma": gamma_list, "beta": beta_list, "optimal value": value, "nfev": nfev}

    def optimize(self, objective_function):
        if self._initial_point is None:
            self._initial_point = np.array([np.random.random() for x in range(2 * self._p)])
        else:
            try:
                if len(self._initial_point) != 2 * self._p:
                    raise ArrayShapeError("The shape of initial parameters is not match with p")
            except ArrayShapeError as e:
                print(e)
                sys.exit()

        return self._minimize(objective_function)


# This code is part of Qcover.
#
# (C) Copyright BAQIS 2021, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import logging
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
import time
from Qcover.applications.common import get_ising_matrix, get_weights_graph, random_number_list


logger = logging.getLogger(__name__)


class Max2Sat:
    """
    max-2-sat problem:
    each clause consists of two literals and a clause is satisfied if either or both literals are true.
    minimizing the number of clauses not satisfied
    """
    def __init__(self,
                 clauses: np.array = None,
                 variable_no: int = None,
                 weight_range: tuple = (1, 100),
                 seed: int = None)-> None:
        """
        Args:
            clauses (np.array): a matrix of clauses, matrix entry is 1 if the literal variable is true, 
            0 if the literal variable is false, -1 for the complement of the variable
            variable_no (int): number of variables
            (require input: clauses)
            
        Returns:
            clauses, variable_no 
        
        Example:
            clauses_matrix = [[1,1,0,0],[1,-1,0,0],[-1,1,0,0],[-1,-1,0,0],[-1,0,1,0],[-1,0,-1,0],
                          [0,1,-1,0],[0,1,0,1],[0,-1,1,0],[0,-1,-1,0],[0,0,1,1],[0,0,-1,-1]]
            variable_number = len(clauses_matrix[0])
        """

        if clauses is None:
            assert variable_no is not None
            
            self._clauses = np.array(clauses, dtype='float64')
            self._variable_no = variable_no
            self._weight_range = weight_range
            # self._seed = seed
            self._clauses = random_number_list(n=self._length,
                                                   weight_range=self._weight_range,
                                                   seed=seed)  #self._seed
        else:
            self._clauses = np.array(clauses, dtype='float64')#clauses
            self._variable_no = variable_no
            self._weight_range = weight_range
            # self._seed = seed

        self._qmatrix = None
        self._shift = None
        
    def get_Qmatrix(self):
        """
        get the Q matrix in QUBO model of number partition problem
        
        Args:
            numbers (np.array): the number set that to be divided
            
        Returns:
        q_mat (np.array): the the Q matrix of QUBO model.

        ..math::
            if a row of self._clauses contains:
            literal x and literal j are both true, f(x)=1-x[i]-x[j]+x[i]x[j]
            literal x is true, the complement of literal j, f(x)+=x[j]-x[i]x[j]
            the complement of literal i, complement of literal j, f(x)+=x[i]x[j]
        """

        set_mat = np.zeros((self._variable_no,self._variable_no))
        q_mat = np.array(set_mat, dtype='float64')
        shift = 0.0

        for a in range(len(self._clauses)): 
           indices_1 = [i for i, x in enumerate(self._clauses[a]) if x == 1.]
           if len(indices_1)>1:
               q_mat[indices_1[0]][indices_1[1]] = 0.5
               q_mat[indices_1[1]][indices_1[0]] = 0.5
               q_mat[indices_1[0]][indices_1[0]] += -1.0
               q_mat[indices_1[1]][indices_1[1]] += -1.0
               shift += 1.0
               
           indices_2 = [i for i, x in enumerate(self._clauses[a]) if x == -1.]
           if len(indices_2)>1:
               q_mat[indices_2[0]][indices_2[1]] += 0.5
               q_mat[indices_2[1]][indices_2[0]] += 0.5
        
           for i in range(self._variable_no):
              for j in range(self._variable_no):
                  if self._clauses[a][i]==1. and self._clauses[a][j]==-1.:
                      indices_3 = [i for i, x in enumerate(self._clauses[a]) if x == 1.]
                      indices_4 = [i for i, x in enumerate(self._clauses[a]) if x == -1.]
                      q_mat[indices_3[0]][indices_4[0]] += -0.5
                      q_mat[indices_4[0]][indices_3[0]] += -0.5
                      q_mat[indices_4[0]][indices_4[0]] += 1.0
           
        return q_mat, shift
        

    def max_2_sat_value(self, x, w):  
        """Compute the value of a max-2-sat problem.

        Args:
            x (numpy.ndarray): binary string as numpy array.
            number_list (numpy.ndarray): list of numbers in the instance.

         Returns:
            float: value of the problem.
        """
       
        if self._qmatrix is None:
            self._qmatrix = self.get_Qmatrix()

        X = np.matmul(x, np.matmul(self._qmatrix, np.transpose(x))) 
        return X

    def run(self):
        if self._qmatrix is None:
            self._qmatrix, self._shift = self.get_Qmatrix()

        qubo_mat = self._qmatrix
        ising_mat = get_ising_matrix(qubo_mat)
        m2s_graph = get_weights_graph(ising_mat)
        return m2s_graph, self._shift
    


# This code is part of Qcover.
#
# (C) Copyright BAQIS 2021, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import logging
import numpy as np
import random
import networkx as nx
import matplotlib.pyplot as plt
import time, os
from numbers import Number
import pandas as pd
from numpy.random import choice
from Qcover.applications.common import get_ising_matrix, get_weights_graph, random_number_list


logger = logging.getLogger(__name__)


class SetPacking:
    """
    Set Packing problem:
    For a finite set S, find a packing (subset of S) that miximises 
    the total weight of the contained objects without violating any constraint.
    
    For the weighted graph, the weight sum of vertices in the subset is maximum.
    """
    def __init__(self,
                 element_list: list = None,
                 length: int = None,
                 weight_range: tuple = (1, 100),
                 element_set: list = None, #constriants from subsets
                 weight: list = None,
                 P: float = None,
                 seed: int = None)-> None:
        """
        Args:
            element_list (list): a list of elements 
            length (int): length of number_list
            element_set (list): a list of constraints
            weight (list): list of weight for element in number_list, default values are 1s
            P (int): the penalty value for the penalty terms
            (require input: element_list, element_set, weight, P)
            
         Returns: 
            element_list, length, weight_range, weight, element_set, constraints, P
            
        Example:
            element_list = ['a','b','c','d']
            element_list_len = len(element_list)
            element_weight = [10,8,10,12]
            subsets = [[1,2],[1,3,4]] 
            penalty = 6
        """
        if element_list is None and element_set is None:
            assert length is not None
            assert len(element_set) is not None
            
            self._length = length
            self._weight_range = weight_range
            self._weight = weight
            # self._weight = np.ones((self._length,), dtype=int)
            self._element_set = element_set
            self._constraints = self.constraints()
            self._P = P
            # self._seed = seed
            self._element_list = random_number_list(n=self._length,
                                                   weight_range=self._weight_range,
                                                   seed=seed) #self._seed

        else:
            for n in [element_list]:
                if isinstance(n, Number) == False:
                    self._element_list = pd.factorize(element_list)[0] 
            self._length = len(self._element_list)
            self._weight_range = (np.abs(self._element_list).min(), np.abs(self._element_list).max())
            # self._seed = seed
            self._weight = weight
            # self._weight = np.ones((self._length,), dtype=int)
            self._element_set = element_set
            self._constraints = self.constraints()
            self._P = P

        self._qmatrix = None
        self._shift = None
            
    @property
    def length(self):
        return self._length

    @property
    def weight(self):
        return self._weight
    
    def constraints(self): 
        """
        If two elements in the list are included in the same constraint, 
        the constraints matrix entry is 1, vice versa
        """
        constraints_matrix = np.zeros((self._length,self._length), dtype='float64')
        for i in range(self._length): #list length
            for j in range(self._length):
                if i!=j:
                    for a in range(len(self._element_set)): #number of constraints
                        if i+1 in self._element_set[a] and j+1 in self._element_set[a]: 
                            constraints_matrix[i][j] = 1.0
                        
        return constraints_matrix

    def update_args(self, length, weight,constraints):
        self._length = length
        self._weight = weight
    
    def get_Qmatrix(self):
        """
        get the Q matrix in QUBO model of set packing problem

        Args:
            numbers (np.array): the number set that to be divided

        Returns:
            q_mat (np.array): the the Q matrix of QUBO model.
            
        ..math::
        minimise x(T)Qx
        Q[i][j] = (P * constraints[i][j])/2 
        Q[i][i] = -weight[i] 
        """
        set_mat = np.eye(self._length)
        q_mat = np.array(set_mat, dtype='float64')
        # q_mat = np.eye(self._length, dtype='float64')
        
        for i in range(self._length):
            q_mat[i][i] = -self._weight[i]
            for j in range(self._length):
                if i==j: 
                    continue
                elif abs(self._constraints[i][j] - 0.) <= 1e-8:
                    q_mat[i][j] = 0.0
                else:
                    q_mat[i][j] = self._P/2.0 
        
        shift = 0.0

        return q_mat, shift

    def set_packing_value(self, x,w): 
        """Compute the value of a partition.

        Args:
            x (numpy.ndarray): binary string as numpy array.
            number_list (numpy.ndarray): list of numbers in the instance.

        Returns:
            float: value of the packing.
        """
        if self._qmatrix is None:
            self._qmatrix = self.get_Qmatrix()

        X = np.matmul(x, np.matmul(self._qmatrix, np.transpose(x))) 
        return X

    def run(self):
        if self._qmatrix is None:
            self._qmatrix, self._shift= self.get_Qmatrix()

        qubo_mat = self._qmatrix
        ising_mat = get_ising_matrix(qubo_mat)
        sp_graph = get_weights_graph(ising_mat)
        return sp_graph, self._shift


# This code is part of Qcover.
#
# (C) Copyright BAQIS 2021, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Object to solve QAOA problems

The QAOA problems can be represented as an Ising model, and then be transformed to a DAG.
The directed acyclic graph is decomposed by a specified p value, and these subgraphs then
be transformed as circuits and be executed on simulators, using optimizer to get the
optimal parameters of the original Ising model
"""

import sys
import time
from itertools import permutations

from typing import Optional
from collections import defaultdict
import numpy as np
import networkx as nx
from Qcover.utils import get_graph_weights
from Qcover.optimizers import Optimizer, COBYLA
from Qcover.backends import Backend, CircuitByQiskit, CircuitByTensor
from Qcover.exceptions import GraphTypeError, UserConfigError
import warnings
warnings.filterwarnings("ignore")


class Qcover:
    """
    Qcover is a QAOA solver
    """
    # pylint: disable=invalid-name
    def __init__(self,
                 graph: nx.Graph = None,
                 p: int = 1,
                 optimizer: Optional[Optimizer] = COBYLA(),
                 backend: Optional[Backend] = CircuitByQiskit(),
                 research_obj: str = "QAOA"
                 ) -> None:

        assert graph is not None
        self._simple_graph = graph
        self._p = p
        self._research_obj = research_obj
        self._backend = backend
        self._backend._p = p
        self._backend._origin_graph = graph
        self._backend._research = research_obj
        self._optimizer = optimizer
        self._optimizer._p = p

    @property
    def p(self):
        return self._p

    @p.setter
    def p(self, ap):
        self._p = ap

    @property
    def backend(self):
        return self._backend

    @property
    def optimizer(self):
        return self._optimizer

    @property
    def simple_graph(self):
        return self._simple_graph

    @simple_graph.setter
    def simple_graph(self, graph):
        """
        according to the type of graph(nx.graph / tuple) to set the value of
        self._simple_graph
        """
        if isinstance(graph, nx.Graph):
            self._simple_graph = graph
        elif isinstance(graph, tuple):
            assert (len(graph) >= 2) and (len(graph) <= 3)

            if len(graph) == 2:
                node_num, edge_num = graph
                wr = None
            elif len(graph) == 3:
                node_num, edge_num, wr = graph

            nodes, edges = Qcover.generate_graph_data(node_num, edge_num, wr)
            self._simple_graph = self.generate_weighted_graph(nodes, edges)
        elif isinstance(graph, list):
            assert len(graph) == 3
            node_list, edge_list, weight_range = graph
            self._simple_graph = Qcover.generate_weighted_graph(node_list, edge_list, weight_range)
        else:
            print(
                "Error: the argument graph should be a instance of nx.Graph or a tuple formed as (node_num, edge_num)")

    @staticmethod
    def generate_graph_data(node_num, edge_num, weight_range=10):
        """
        generate a simple graph‘s weights of nodes and edges with
        node number is node_num, edge number is edge_num
        Args:
            node_num (int): node number in graph
            edge_num (int): edge number in graph
            weight_range (int): weight range of nodes and edges
        Return:
            nodes(set of tuple(nid, node_weight)), edges(set of tuple(nid1, nid2, edge_weight))
        """
        if weight_range is None:
            weight_range = 10

        nodes = set()
        for i in range(node_num):
            ndw = np.random.choice(range(weight_range))
            nodes |= {(i, ndw)}

        edges = set()
        cnt = edge_num
        max_edges = node_num * (node_num - 1) / 2
        if cnt > max_edges:
            cnt = max_edges
        while cnt > 0:
            u = np.random.randint(node_num)
            v = np.random.randint(node_num)
            if u == v:  # without self loop
                continue
            flg = 0
            for e in edges:  # without duplicated edges
                if set(e[:2]) == set([v, u]):
                    flg = 1
                    break
            if flg == 1:
                continue
            edw = np.random.choice(range(weight_range))
            edges |= {(u, v, edw)}
            cnt -= 1
        return nodes, edges

    @classmethod
    def generate_weighted_graph(cls, nodes, edges, weight_range=10):
        """
        generate graph from nodes list and edges list which identify the nodes and edges
        that should be add in graph, and the random weight range of every node and edge.
        Args:
            nodes (list/set): list of node idex / node-weight map, element form is tuple(nid, weight)
            edges (list/set): list of edge: (e_idex1, e_idex2) / edge-weight map, element form is tuple(nid1, nid2, edge_weight)
            weight_range (int): random weight range of every node and edge
        Returns:
            g (nx.Graph): graph generated by args
        """
        g = nx.Graph()
        if isinstance(nodes, list) and isinstance(edges, list):
            for v in nodes:
                w = np.random.choice(range(weight_range))
                g.add_node(v, weight=w)

            for e in edges:
                w = np.random.choice(range(weight_range))
                g.add_edge(e[0], e[1], weight=w)
        else:
            for item in nodes:
                g.add_node(item[0], weight=item[1])

            for item in edges:
                g.add_edge(item[0], item[1], weight=item[2])
        return g

    # @staticmethod
    # def get_graph_weights(graph):
    #     """
    #     get the weights of nodes and edges in graph
    #     Args:
    #         graph (nx.Graph): graph to get weight of nodes and edges
    #     Return:
    #         node weights form is dict{nid1: node_weight}, edges weights form is dict{(nid1, nid2): edge_weight}
    #     """
    #     nodew = nx.get_node_attributes(graph, 'weight')
    #     edw = nx.get_edge_attributes(graph, 'weight')
    #     edgew = edw.copy()
    #     for key, val in edw.items():
    #         edgew[(key[1], key[0])] = val
    #
    #     return nodew, edgew

    def generate_subgraph(self, graph, dtype: str, p):
        """
        according to the arguments of dtype and p to generate subgraphs from graph
        Args:
            graph (nx.Graph): graph to be decomposed
            dtype (string): set "node" or "edge", the ways according to which to decompose the graph
            p (int): the p of subgraphs
        Return:
            subg_dict (dict) form as {node_id : subg, ..., (node_id1, node_id2) : subg, ...}
        """
        if dtype not in ["node", "edge"]:
            print("Error: wrong dtype, dtype should be node or edge")
            return None

        nodes_weight, edges_weight = get_graph_weights(graph)

        subg_dict = defaultdict(list)
        if dtype == 'node':
            for node in graph.nodes:
                node_set = {(node, nodes_weight[node])}
                edge_set = set()
                for i in range(p):
                    new_nodes = {(nd2, nodes_weight[nd2]) for nd1 in node_set for nd2 in graph[nd1[0]]}
                    new_edges = {(nd1[0], nd2, edges_weight[nd1[0], nd2]) for nd1 in node_set for nd2 in graph[nd1[0]]}
                    node_set |= new_nodes
                    edge_set |= new_edges

                subg = self.generate_weighted_graph(node_set, edge_set)
                subg_dict[node] = subg
        else:
            for edge in graph.edges:
                node_set = {(edge[0], nodes_weight[edge[0]]), (edge[1], nodes_weight[edge[1]])}
                edge_set = {(edge[0], edge[1], edges_weight[edge[0], edge[1]])}

                for i in range(p):
                    new_nodes = {(nd2, nodes_weight[nd2]) for nd1 in node_set for nd2 in graph[nd1[0]]}
                    new_edges = {(nd1[0], nd2, edges_weight[nd1[0], nd2]) for nd1 in node_set for nd2 in
                                 graph.adj[nd1[0]]}
                    node_set |= new_nodes
                    edge_set |= new_edges

                subg = self.generate_weighted_graph(node_set, edge_set)
                subg_dict[edge] = subg
        return subg_dict

    def graph_decomposition(self, graph, p):
        """
        according to dtype to decompose graph
        Args:
            graph (nx.Graph): graph to be composed
            p (int): the p of subgraphs
        """
        if p <= 0:
            warnings.warn(" the argument of p should be >= 1 in qaoa problem, "
                          "so p would be set to the default value at 1")
            p = 1

        subg_node = self.generate_subgraph(graph, "node", p)
        subg_edge = self.generate_subgraph(graph, "edge", p)
        element_to_graph = {}
        for k, v in subg_node.items():
            element_to_graph[k] = v

        for k, v in subg_edge.items():
            element_to_graph[k] = v
        return element_to_graph

    @staticmethod
    def solve_basic_graph(graph: nx.Graph):
        nodes = graph.nodes
        min_H = np.inf
        basic_sol = dict()
        for i in range(2 ** len(nodes)):
            cur_H = 0
            tmp_sol = dict()
            binary_list = [int(x) for x in bin(i)[2:].zfill(len(nodes))]
            for i, nd in enumerate(nodes):
                tmp_sol[nd] = binary_list[i]

            for nd in graph.nodes:
                sigma_nd = 1 if tmp_sol[nd] == 1 else -1
                cur_H += graph.nodes[nd]["weight"] * sigma_nd

            for ed in graph.edges:
                u, v = ed
                sigma_u = 1 if tmp_sol[u] == 1 else -1
                sigma_v = 1 if tmp_sol[v] == 1 else -1
                cur_H += graph.adj[u][v]["weight"] * sigma_u * sigma_v

            if cur_H < min_H:
                min_H = cur_H
                basic_sol = tmp_sol
        return basic_sol

    @staticmethod
    def get_solution(graph, basic_sol: dict = None):
        sol = defaultdict(lambda: -1)
        queue = list(basic_sol.keys()) if basic_sol is not None else []
        for i in graph.nodes:
            if sol[i] == -1:  # len(graph.neighbors(i)) != 0 and
                queue.append(i)
                while len(queue) > 0:
                    u = queue[0]
                    queue.pop(0)
                    if sol[u] == -1:
                        sol[u] = 0
                    if u not in graph:
                        continue
                    neighbor_u = list(graph.neighbors(u))
                    for v in neighbor_u:
                        if sol[v] == -1:
                            sol[v] = sol[u] if graph.adj[u][v]["weight"] > 0 else 1 - sol[u]
                            queue.append(v)
        return sol

    def calculate(self, pargs, p=None):
        """
        The framework function which use the backend to calculate the value of expectation,
        and be used as the object function in the optimization function of the optimizer
        Args:
            pargs: the value of the parameter alpha and beta in the circuit
            p: the integer used to define the number of layers the current circuit needs to be superimposed
        Returns:
            the value of expectation calculated by backends
        """
        p = self._p if p is None else p
        element_to_graph = self.graph_decomposition(graph=self._simple_graph, p=p)
        # element_to_graph = {(0, 1):self._simple_graph}

        # checking graph type of given problem
        if not isinstance(self._backend, CircuitByTensor):
            for k, v in element_to_graph.items():
                ncnt, ecnt = len(v.nodes), len(v.edges)
                try:
                    nreq1 = ncnt * (ncnt - 1) <= 2 * ecnt and ncnt >= 20
                    nreq2 = isinstance(k, int) and v.degree[k] >= 30
                    if nreq1 or nreq2:
                        raise GraphTypeError("The problem is transformed into a dense graph, " \
                           "which is difficult to be solved effectively by Qcover")
                except GraphTypeError as e:
                    print(e)
                    # if not self._hard_to_calcute:
                    #     print(e)
                    #     self._hard_to_calcute = True
                    # sys.exit()

        self._backend._pargs = pargs
        self._backend._element_to_graph = element_to_graph
        return self._backend.expectation_calculation(p)

    def run_qaoa(self):
        """
        run Qcover code to solve the given problem
        Args:
            node_num: nodes number in the graphical representation of the problem
            edge_num: edges number in the graphical representation of the problem
            is_parallel: run programs in parallel or not

        Returns:
            results of the problem, which including the optimal parameters of the circuit model,
            the optimal expectation value and the number of times the optimizer iterates
        """
        # nodes_weight, edges_weight = get_graph_weights(self._simple_graph)
        # self._backend._nodes_weight = nodes_weight
        # self._backend._edges_weight = edges_weight

        x, fun, nfev = self._optimizer.optimize(objective_function=self.calculate)
        res = {"Optimal parameter value": x, "Expectation of Hamiltonian": fun, "Total iterations": nfev}
        return res

    def run_rqaoa(self, node_threshold, iter_time, corr_method):
        try:
            if iter_time < 1:
                raise UserConfigError("iter_time should be a value greater than 1")
        except UserConfigError as e:
            print(e)
            print("iter_time will be set to 1")
            iter_time = 1

        sexp = []

        node_sol = len(self._simple_graph)
        current_g = self._simple_graph
        node_num = len(current_g.nodes)
        pathg = nx.Graph()
        while node_num > node_threshold:
            # nodes_weight, edges_weight = get_graph_weights(current_g)
            # self._backend._nodes_weight = nodes_weight
            # self._backend._edges_weight = edges_weight
            self._backend._origin_graph = current_g

            optization_rounds = iter_time
            while optization_rounds > 0:
                self._optimizer.optimize(objective_function=self.calculate)  # x, fun, nfev =

                if corr_method == 'expectation':
                    # origin format of RQAOA
                    exp_sorted = sorted(self._backend.element_expectation.items(),
                                        key=lambda item: abs(item[1]),
                                        reverse=True)
                else:
                    corr = dict()
                    for key, val in self._backend.element_expectation.items():
                        if isinstance(key, tuple):
                            vi, vj = self._backend.element_expectation[key[0]], self._backend.element_expectation[key[1]]
                            corr[key] = (val - vi * vj) / (1 - vi * vj)
                        else:
                            continue
                    exp_sorted = sorted(corr.items(), key=lambda item: abs(item[1]), reverse=True)

                u, v = exp_sorted[0][0]
                exp_value = abs(exp_sorted[0][1])

                # print("iteration on %d, max expectation is %lf" % (optization_rounds, exp_value))
                # print("--------------------------------------")
                if exp_value - 1e-8 >= 0.5:
                    break
                optization_rounds -= 1

            # if exp_value < 0.5:  # if the relation between two node is small, then run search method
            #     break
            sexp.append(exp_value)
            correlation = 1 if exp_sorted[0][1] - 1e-8 > 0 else -1
            pathg.add_edge(u, v, weight=correlation)
            if u > v:
                u, v = v, u

            for nd in current_g.neighbors(v):
                if nd == u:
                    continue
                if nd not in current_g.neighbors(u):
                    current_g.add_edge(u, nd, weight=correlation * current_g.adj[v][nd]["weight"])
                else:
                    current_g.adj[u][nd]["weight"] += correlation * current_g.adj[v][nd]["weight"]
                    # current_g.adj[nd][u]["weight"] = current_g.adj[u][nd]["weight"]
            current_g.remove_node(v)
            node_num = len(current_g.nodes)

        basic_sol = None
        if len(current_g.nodes) > 1:   #node_threshold
            basic_sol = self.solve_basic_graph(current_g)

        if basic_sol is None or len(basic_sol) < node_sol:
            tmp_sol = self.get_solution(pathg, basic_sol)
            if basic_sol is None:
                solution = tmp_sol
            else:
                solution = {**tmp_sol, **basic_sol}
        else:
            solution = basic_sol
        return solution, sexp

    def run(self,
            node_num=None, edge_num=None,
            is_parallel=False,
            mode='QAQA',
            node_threshold=1,
            iter_time=1,
            corr_method='expectation'):

        if self._simple_graph is None:
            if node_num is None or edge_num is None:
                print("Error: the graph to be solved is not specified, "
                      "arguments of form (node_number, edge_number) should be given")
                return None
            self.simple_graph(node_num, edge_num)

        self._backend._is_parallel = is_parallel
        if mode == 'QAQA':
            res = self.run_qaoa()
        else:
            res = self.run_rqaoa(node_threshold, iter_time, corr_method)

        return res


# usage example
if __name__ == '__main__':

    p = 1
    g = nx.Graph()
    nodes = [(0, 0), (1, 0), (2, 0)]
    edges = [(0, 1, -1), (1, 2, -1)]

    # nodes = [(0, 0), (1, 0), (2, 0), (3, 0), (4, 0)]
    # edges = [(0, 1, -1), (1, 2, -1), (2, 3, -1), (3, 4, -1)]

    for nd in nodes:
        u, w = nd[0], nd[1]
        g.add_node(int(u), weight=int(w))
    for ed in edges:
        u, v, w = ed[0], ed[1], ed[2]
        g.add_edge(int(u), int(v), weight=int(w))

    #Ising test
    # g.add_node(0, weight=0)
    # for i in range(1, 30):
    #     g.add_node(i, weight=0)
    #     g.add_edge(i, i - 1, weight=-1)

    # nx.draw(g)
    # from Qcover.applications import MaxCut
    # mxt = MaxCut(g)
    # mxt = MaxCut(node_num=20, node_degree=3)
    #10 3 0.035  100 3 0.029
    #10 6 0.028  100 6 134.56/ce
    # g, shift = mxt.run()

    from Qcover.optimizers import GradientDescent, SPSA
    optg = GradientDescent(options={'maxiter':300, 'learning_rate':0.05, 'tol':1e-6})
    opta = SPSA(options={'maxiter':300, 'tol':1e-6})
    optc = COBYLA(options={'tol': 1e-3, 'disp': True})
    qiskit_bc = CircuitByQiskit(expectation_calc_method="statevector")
    from Qcover.backends import CircuitByQton, CircuitByQulacs, CircuitByProjectq, CircuitByTensor, CircuitByCirq
    qulacs_bc = CircuitByQulacs()
    # cirq_bc = CircuitByCirq()
    # projectq_bc = CircuitByProjectq()
    # ts = CircuitByTensor()
    # qt = CircuitByQton()  #expectation_calc_method="tensor"

    qc = Qcover(g, p,
                # research_obj="QAOA",
                optimizer=opta,  #@ optc,
                backend=qiskit_bc)  #, qt, qulacs_bc, cirq_bc, projectq_bc

    st = time.time()
    sol = qc.run(is_parallel=False, mode='QAQA')  #True
    ed = time.time()
    print("time cost by QAOA is:", ed - st)
    print("solution is:", sol)
    params = sol["Optimal parameter value"]
    qc.backend._pargs = params
    out_count = qc.backend.get_result_counts(params, g)
    res_exp = qc.backend.expectation_calculation()
    print("the optimal expectation is: ", res_exp)
    import matplotlib.pyplot as plt
    from qiskit.visualization import plot_histogram
    plot_histogram(out_count)
    plt.show()

    # test RQAOA
    # import matplotlib.pyplot as plt
    # from Qcover.applications import MaxCut
    #
    # p = 1
    # node_num = [10, 50, 100, 500, 1000]
    # node_d = [3, 4, 5, 6]
    # for i in node_num:
    #     et, gt = [], []
    #     ev, gv = [], []
    #     for nd in node_d:
    #         print("node number is %d, node degree is %d" % (i, nd))
    #         print("----------------------------------------------")
    #         mxt = MaxCut(node_num=i, node_degree=nd)
    #         g, shift = mxt.run()
    #
    #         g_exp = g.copy()
    #         qc_exp = Qcover(g_exp, p,
    #                     optimizer=COBYLA(options={'tol': 1e-3, 'disp': True}),
    #                     backend=CircuitByQiskit(expectation_calc_method="statevector"))  # qt, ,qulacs_bc
    #
    #         st = time.time()
    #         sol_exp, sexp_exp = qc_exp.run(mode='RQAQA', node_threshold=1, iter_time=3, corr_method='expectation')
    #         ed = time.time()
    #         t1 = ed - st
    #
    #         g_g = g.copy()
    #         qc_g = Qcover(g_g, p,
    #                     optimizer=COBYLA(options={'tol': 1e-3, 'disp': True}),
    #                     backend=CircuitByQiskit(expectation_calc_method="statevector"))  # qt, ,qulacs_bc
    #         st = time.time()
    #         sol_g, sexp_g = qc_g.run(mode='RQAQA', node_threshold=1, iter_time=3, corr_method='g')
    #         ed = time.time()
    #         t2 = ed - st
    #
    #         exp_e, exp_g = 0, 0
    #         for (x, y) in g.nodes.data('weight', default=0):
    #             exp_e += y * (sol_exp[x] * 2 - 1)
    #             exp_g += y * (sol_g[x] * 2 - 1)
    #         for (u, v, c) in g.edges.data('weight', default=0):
    #             exp_e += c * (sol_exp[u] * 2 - 1) * (sol_exp[v] * 2 - 1)
    #             exp_g += c * (sol_g[u] * 2 - 1) * (sol_g[v] * 2 - 1)
    #
    #         et.append(t1)
    #         gt.append(t2)
    #
    #         ev.append(exp_e)
    #         gv.append(exp_g)
    #
    #         plt.figure(0)
    #         plt.plot(range(len(sexp_exp)), sexp_exp, "ob-", label="origin")
    #         plt.plot(range(len(sexp_g)), sexp_g, "^r-", label="correlation")
    #         plt.ylabel('expectation')
    #         plt.xlabel('iteration rounds')
    #         plt.title("node is %d, degree is %d" % (i, nd))
    #         plt.legend()
            # plt.savefig('/home/baqis/Qcover/result_log/correlation_exp/iter_%d_%d.png' % (i, nd))
            # plt.savefig('E:/Working_projects/QAOA/Qcover/result_log/correlation_exp/iter_%d_%d.png' % (i, nd))
        #     plt.savefig('/home/puyanan/Qcover/result_log/correlation_exp/iter_%d_%d.png' % (i, nd))
        #     plt.close('all')
        #
        # plt.figure(1)
        # plt.plot(node_d, et, "ob-", label="origin")
        # plt.plot(node_d, gt, "^r-", label="correlation")
        # plt.ylabel('Time cost')
        # plt.xlabel('node degree')
        # plt.title("node is %d" % i)
        # plt.legend()
        # plt.savefig('/home/baqis/Qcover/result_log/correlation_exp/time_cost_%d.png' % i)
        # plt.savefig('E:/Working_projects/QAOA/Qcover/result_log/correlation_exp/time_cost_%d.png' % i)
        # plt.savefig('/home/puyanan/Qcover/result_log/correlation_exp/time_cost_%d.png' % i)
        #
        # plt.figure(2)
        # plt.plot(node_d, ev, "*g-", label="origin")
        # plt.plot(node_d, gv, "dy-", label="correlation")
        # plt.ylabel('expectation of hamitonian')
        # plt.xlabel('node degree')
        # plt.title("node is %d" % i)
        # plt.legend()
        # plt.savefig('/home/baqis/Qcover/result_log/correlation_exp/hamitonian_exp_%d.png' % i)
        # plt.savefig('E:/Working_projects/QAOA/Qcover/result_log/correlation_exp/hamitonian_exp_%d.png' % i)
        # plt.savefig('/home/puyanan/Qcover/result_log/correlation_exp/hamitonian_exp_%d.png' % i)
        #
        # plt.close('all')




# print("time cost by RQAOA with expectation is:", t1)
# print("expectation list with expectation:", exp_e)
# print("solution of expectation is:", sol_exp)
# # {8: 0, 4: 1, 2: 0, 3: 1, 7: 0, 5: 1, 0: 0, 6: 0, 1: 1, 9: 1})
#
# print("time cost by RQAOA with correlation is:", t2)
# print("expectation list with correlation:", exp_g)
# print("solution of correlation is:", sol_g)







import networkx as nx

def get_graph_weights(graph):
    """
    get the weights of nodes and edges in graph
    Args:
        graph (nx.Graph): graph to get weight of nodes and edges
    Return:
        node weights form is dict{nid1: node_weight}, edges weights form is dict{(nid1, nid2): edge_weight}
    """
    nodew = nx.get_node_attributes(graph, 'weight')
    edw = nx.get_edge_attributes(graph, 'weight')
    edgew = edw.copy()
    for key, val in edw.items():
        edgew[(key[1], key[0])] = val

    return nodew, edgew


# This code is part of Qcover.
#
# (C) Copyright BAQIS 2021, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import os
import sys
import logging
import time
import numpy as np
import networkx as nx
import random
import matplotlib.pyplot as plt
from Qcover.applications.common import get_ising_matrix, get_weights_graph, random_number_list #Qcover.applications.


logger = logging.getLogger(__name__)


class NumberPartition:
    """
    number partition problem:
    given the set of numbers, divided it into two parties
    so the sum of numbers in two parties are as close as possible
    """
    def __init__(self,
                 number_list: np.array = None,
                 length: int = None,
                 weight_range: tuple = (1, 100),
                 seed: int = None) -> None:

        if number_list is None:
            assert length is not None

            self._length = length
            self._weight_range = weight_range
            self._seed = seed
            self._number_list = random_number_list(n=self._length,
                                                   weight_range=self._weight_range,
                                                   seed=self._seed)
        else:
            self._number_list = number_list
            self._length = len(number_list)
            self._weight_range = (np.abs(number_list).min(), np.abs(number_list).max())
            self._seed = seed
            
        self._qmatrix = None
        self._shift = None
        
    @property
    def length(self):
        return self._length

    @property
    def weight_range(self):
        return self._weight_range

    @property
    def number_list(self):
        return self._number_list

    def update_args(self, length, weight_range):
        self._length = length
        self._weight_range = weight_range

        self._number_list = random_number_list(n=self._length, weight_range=self._weight_range, seed=self._seed)   # self.get_random_number_list()

    def get_Qmatrix(self):
        """
        get the Q matrix in QUBO model of number partition problem

        Args:
            numbers (np.array): the number set that to be divided

        Returns:
            q_mat (np.array): the the Q matrix of QUBO model.
        """
        
        all_sum = np.sum(self._number_list)
        q_mat = np.eye(self._length)

        for i in range(self._length):
            q_mat[i][i] = self._number_list[i] * (self._number_list[i] - all_sum)
            for j in range(self._length):
                if i == j:
                    continue
                q_mat[i][j] = self._number_list[i] * self._number_list[j]
                
        shift = 0.0
        
        return q_mat,shift
    
    def partition_value(self, x, number_list):  #
        """Compute the value of a partition.

        Args:
            x (numpy.ndarray): binary string as numpy array.
            number_list (numpy.ndarray): list of numbers in the instance.

        Returns:
            float: difference squared between the two sides of the number
                partition.
        """
        diff = np.sum(number_list[x == 0]) - np.sum(number_list[x == 1])
        return diff * diff

    def run(self):
        if self._qmatrix is None:
            self._qmatrix, self._shift = self.get_Qmatrix()
            
        qubo_mat = self._qmatrix
        ising_mat = get_ising_matrix(qubo_mat)
        np_graph = get_weights_graph(ising_mat)
        return np_graph, self._shift


import sys
import numpy as np
from typing import Optional
from scipy import optimize as opt
from math import *
from Qcover.optimizers import Optimizer
from Qcover.exceptions import ArrayShapeError


class Fourier(Optimizer):
    """
        Fourier optimizer: a heuristic optimization method for QAOA,
        implemented according to the paper
        "Quantum Approximate Optimization Algorithm: Performance, Mechanism, and Implementation on Near-Term Devices"
    """
    def __init__(self,
                 p: int = 1,
                 q: Optional[int] = None,   # 4
                 r: Optional[int] = 0,
                 alpha: Optional[float] = 0.6,
                 optimize_method: Optional[str] = 'COBYLA',
                 options: dict = None, #{'maxiter':300, 'disp':True, 'rhobeg': 1.0, 'tol':1e-6},
                 initial_point: Optional[list] = None
                 ) -> None:
        """
        initialize a optimizer of FOURIER with parameters q and R
        Args:
            p: the parameter in QAOA paper
            q: the maximum frequency component allowed in the amplitude parameters <⃗u, ⃗v>
            r: the number of random perturbations to add
            alpha:
        """
        super().__init__()
        self._p = p
        self._q = q if q is not None and (q < self._p and q >= 1) else self._p
        self._r = r
        self._alpha = alpha
        self._optimize_method = optimize_method
        self._options = options
        self._initial_point = initial_point # used to initialize (u, v) that shape are defined by q

        self._objective_function = None

    @property
    def q(self):
        return self._q

    @q.setter
    def q(self, aq):
        self._q = aq if (aq < self._p and aq >= 1) else self._p

    @property
    def r(self):
        return self._r

    @r.setter
    def r(self, ar):
        self._r = ar

    def calculate_gb(self, step, pargs):
        upb = min(self._q, step)
        u, v = pargs[:upb], pargs[upb:]
        gamma, beta = np.zeros(step), np.zeros(step)
        for i in range(1, step + 1):
            for k in range(1, upb + 1):
                gamma[i - 1] += u[k - 1] * sin((k - 0.5) * (i - 0.5) * pi / step)
                beta[i - 1] += v[k - 1] * cos((k - 0.5) * (i - 0.5) * pi / step)

            if gamma[i - 1] < -np.pi / 2:
                gamma[i - 1] = -np.pi / 2 + 0.01
            if beta[i - 1] < -np.pi / 4:
                beta[i - 1] = -np.pi / 4 + 0.01
            if gamma[i - 1] > np.pi / 2:
                gamma[i - 1] = np.pi / 2 - 0.01
            if beta[i - 1] > np.pi / 4:
                beta[i - 1] = np.pi / 4 - 0.01
        return gamma, beta

    def loss_function(self, pargs, step):
        gamma, beta = self.calculate_gb(step, pargs)
        return self._objective_function(np.append(gamma, beta), step)

    def _minimize(self, objective_function):
        """
        minimize the loss function
        Args:
            loss: the loss function

        Returns:
            x: the optimized gamma and beta
            value: the optimized value of loss function
            nfev: is the number of objective function calls
        """
        self._objective_function = objective_function

        nfev = 0
        ul, vl = None, None
        u_best, v_best = None, None
        for j in range(1, self._p + 1):
            u_list, v_list = [], []
            min_val = float("inf")
            if j == 1:
                # Attention:
                # though only used in j=1, but to be consistent with other optimizers,
                # self._initial_point should be defined according to q,
                # which is different with other optimizes
                ul, vl = list(self._initial_point[: j]), list(self._initial_point[self._q: self._q+j])
            else:
                if j <= self._q:
                    ul.append(0)
                    vl.append(0)

                for r in range(self._r + 1):
                    u_nx, v_nx = u_best.copy(), v_best.copy()
                    if r > 0:
                        for i, _ in enumerate(u_best):
                            u_nx[i] = u_nx[i] + self._alpha * np.random.normal(loc=0, scale=fabs(u_best[i]))
                            v_nx[i] = v_nx[i] + self._alpha * np.random.normal(loc=0, scale=fabs(v_best[i]))
                    if j <= self._q:
                        u_nx.append(0)
                        v_nx.append(0)
                    u_list.append(u_nx)
                    v_list.append(v_nx)

            for idx in range(len(u_list) + 1):
                if idx == 0:
                    u_cal, v_cal = ul, vl
                else:
                    u_cal, v_cal = u_list[idx - 1], v_list[idx - 1]

                res = opt.minimize(self.loss_function,
                                    x0=np.append(u_cal, v_cal),
                                    args=j,
                                    method=self._optimize_method,
                                    jac=opt.rosen_der,
                                    options=self._options)

                upb = min(self._q, j)
                if idx == 0:
                    ul, vl = list(res["x"][:upb]), list(res["x"][upb:])

                func_val = res["fun"]
                if func_val < min_val:
                    min_val = func_val
                    u_best, v_best = list(res["x"][:upb]), list(res["x"][upb:])

                nfev += res["nfev"]

        gamma_list, beta_list = self.calculate_gb(self._p, u_best + v_best)
        return np.append(gamma_list, beta_list), min_val, nfev
        # return {"gamma": gamma_list, "beta": beta_list, "optimal value": min_val, "nfev": nfev}

    def optimize(self, objective_function):
        if self._initial_point is None:
            self._initial_point = np.array([np.random.random() for x in range(2 * self._q)])
        else:
            try:
                if len(self._initial_point) != 2 * self._q:
                    raise ArrayShapeError("The shape of initial parameters is not match with q")
            except ArrayShapeError as e:
                print(e)
                sys.exit()

        return self._minimize(objective_function)



# This code is part of Qcover.
#
# (C) Copyright BAQIS 2021, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import logging
import numpy as np
import random
import networkx as nx
import matplotlib.pyplot as plt
import time, os
from numbers import Number
import pandas as pd
from numpy.random import choice
import math
from Qcover.applications.common import get_ising_matrix, get_weights_graph, random_number_list


logger = logging.getLogger(__name__)


class QadraticAssignment:
    """
    Qadratic Assignment problem:
    Given n facilities and n locations along with a flow matrix (f_ij) 
    denoting the flow of material between facilities i and j.
    A distance matrix (d_ij) specifies the distance between sites i and j. 
    The optimization problem is to find an assignment of facilities to locations 
    to minimize the weighted flow across the system. 
    """
    def __init__(self,
                 flow: np.array = None,
                 distance: np.array = None,
                 n: int = None,
                 weight_range: tuple = (1, 100),
                 element_set: np.array = None, #constriants from subsets
                 P: float = None,
                 seed: int = None):
        """
        Args:
            flow (np.array): flow matrix, denoting the flow of material between facilities
            distance (np.array): distance matrix, specifies the distance between sites
            n (int): number of sites = number of facilities
            element_set (np.array): matrix of coefficient of constraints
            P (int): the penalty value for the penalty terms
            (require input: flow, distance, element_set, P)
        
        Returns:
            flow, distance, n, element_set, P
            
        Example:
            flow_matrix = [[0,5,2],[5,0,3],[2,3,0]] 
            distance_matrix = [[0,8,15],[8,0,13],[15,13,0]]
            n = len(flow_matrix)
            subsets = [[1,2,3],[4,5,6],[7,8,9],[1,4,7],[2,5,8],[3,6,9]]
            penalty = 200
        """
        
                
        if flow is None and element_set is None: #and distance is None
            assert n is not None
            assert len(element_set) is not None
            
            self._distance = distance
            self._n = n
            self._weight_range = weight_range
            self._element_set = element_set
            self._P = P
            self._seed = seed
            self._flow = np.array(random_number_list(n=self._n**2,
                                                   weight_range=self._weight_range,
                                                   seed=self._seed)).reshpae(self._n,self._n)

        else:
            self._flow = flow
            self._distance = distance
            self._n = len(flow)
            self._seed = seed
            self._element_set = element_set
            self._P = P

        self._qmatrix = None
        self._shift = None
                    
    # def penalty(self):
    #    """
    #    penalty: some percentage (75% to 150%) of the estimate of the original objective function value
       
    #    Args:
    #        x (numpy.ndarray): binary string as numpy array.
           
    #    Returns:
    #        the penalty value
    #    """
    #    # a = choice([0,1], size = [1,self.length], p=[0.5, 0.5]) 

    #    # estimation = np.dot(a, self._weight).item()*1.5
    #    # return estimation
    #    penalty = 200
    #    return penalty
    
    def get_Qmatrix(self):
        """
        get the Q matrix in QUBO model of set packing problem

        Args:
            numbers (np.array): the number set that to be divided

        Returns:
            q_mat (np.array): the the Q matrix of QUBO model.
        
        ..math::
            minimise x(T)Qx
            Q[(j-1)*n+i][(i-1)*n+j] = flow[j][i] * distance[i][j] + P
            Q[i][i] = -2 * P
        """
        matrix_dimension = self._n * self._n
        q_mat = -2.0 * self._P * np.eye(matrix_dimension, dtype='float64')
        for i in range(matrix_dimension):
           for j in range(matrix_dimension):
               x, y = i // self._n, j // self._n 
               idl, idr = x * self._n, (x + 1.0) * self._n - 1.0 
               if (i >= idl and i <= idr) and (j >= idl and j <= idr):
                   if i == j:
                       q_mat[i][i] = -2.0 * self._P
                   else:
                       q_mat[i][j] = self._P
               else:
                   for l in range(self._n):
                       for k in range(self._n):
                            q_mat[x*self._n+k][y*self._n+l] = self._flow[x][y]*self._distance[k][l] 
                            if self._flow[x][y]*self._distance[k][l] == 0.0:
                                q_mat[x*self._n+k][y*self._n+l] = self._P
        
        shift = 2.0 * self._P
               
        return q_mat, shift

    def quadratic_assignment_value(self, x,w): 
        """Compute the value of a quadratic assignment.

        Args:
            x (numpy.ndarray): binary string as numpy array.
            number_list (numpy.ndarray): list of numbers in the instance.

        Returns:
            float: value of the assignment.
        """
        if self._qmatrix is None:
            self._qmatrix = self.get_Qmatrix()

        X = np.matmul(x, np.matmul(self._qmatrix, np.transpose(x))) 
        return X

    def run(self):
        if self._qmatrix is None:
            self._qmatrix, self._shift= self.get_Qmatrix()

        qubo_mat = self._qmatrix
        ising_mat = get_ising_matrix(qubo_mat)
        qa_graph = get_weights_graph(ising_mat)
        return qa_graph, self._shift


import numpy as np
import networkx as nx
from typing import Optional
from collections import defaultdict

from Qcover.optimizers import Optimizer, COBYLA
from Qcover.backends import Backend, CircuitByQiskit
from Qcover.core import Qcover
import warnings
warnings.filterwarnings("ignore")


class GHZ_Generate:

    def __init__(self,
                 node_num: int,
                 p: int = 1,
                 graph: nx.Graph = None,
                 optimizer: Optional[Optimizer] = COBYLA(),
                 backend: Optional[Backend] = CircuitByQiskit(),
                 ) -> None:
        self._p = p
        self._node_num = node_num
        # self._backend = backend
        # self._optimizer = optimizer

        if graph is None:
            self._g = self.get_graph()
        else:
            if self._node_num != len(graph):
                print("Error: node number should be same with the one in graph to initialize")
                return
            self._g = graph

        self._qc = Qcover(self._g,
                          self._p,
                          optimizer=optimizer,
                          backend=backend,
                          research_obj="GHZ")

    @property
    def p(self):
        return self._p

    @p.setter
    def p(self, ap):
        self._p = ap

    @property
    def node_num(self):
        return self._node_num

    @node_num.setter
    def node_num(self, nd):
        self._node_num = nd

    @property
    def qc(self):
        return self._qc

    def get_graph(self):
        g = nx.Graph()
        g.add_node(0, weight=0)
        for i in range(1, 2 * self._node_num + 1):
            g.add_node(i, weight=0)
            g.add_edge(i - 1, i, weight=-1)
        return g

    def run(self, is_parallel=False, mode='QAQA'):
        if self._g is None:
            self._g = self.get_graph()

        # nx.draw(self._g)
        # if self._qc is None:
        #     self._qc = Qcover(self._g,
        #                       self._p,
        #                       optimizer=self._optimizer,
        #                       backend=self._backend,
        #                       research_obj="GHZ")

        sol = self._qc.run(is_parallel=is_parallel, mode=mode)  # True
        return sol


if __name__ == '__main__':
    p = 1
    node_num = 2

    opti_params = [-0.23533129, -0.51984872, -0.63338606, -0.68489186, -0.71214422,
       -0.74310004, -0.75405114, -0.75529699, -0.75482341, -0.73029963,
       -0.78895398, -0.86082996,  0.73003181,  0.71178062,  0.72944497,
        0.73066588,  0.73455175,  0.7256124 ,  0.71018419,  0.68513461,
        0.66249872,  0.55629759,  0.37117664,  0.11451713]

    from Qcover.optimizers import COBYLA, Interp, SLSQP, L_BFGS_B, GradientDescent, SPSA
    optc = COBYLA(options={'maxiter': 300, 'disp': True, 'rhobeg': 1.0, 'tol': 1e-12}) #, initial_point=opti_params
    opti = Interp(optimize_method="COBYLA", options={'tol': 1e-12, 'disp': False}) #, initial_point=opti_params, "optimal_value": -2*node_num, "approximate_ratio": 0.9
    opts = SLSQP(options={'maxiter': 100,
                          'ftol': 1e-06,
                          'iprint': 1,
                          'disp': False,
                          'eps': 1.4901161193847656e-08,
                          'finite_diff_rel_step': None})

    optl = L_BFGS_B(options={'disp': None,
                             'maxcor': 10,
                             'ftol': 2.220446049250313e-09,
                             'gtol': 1e-05,
                             'eps': 1e-08,
                             'maxfun': 15000,
                             'maxiter': 15000,
                             'iprint': -1,
                             'maxls': 20,
                             'finite_diff_rel_step': None})  #

    optg = GradientDescent(options={'maxiter':300, 'learning_rate':0.05, 'tol':1e-6})
    opta = SPSA(options={'A':34, 'R': 15, 'maxiter':300, 'tol':1e-6})
    from Qcover.backends import CircuitByQiskit, CircuitByQton, CircuitByQulacs, CircuitByProjectq, CircuitByTensor, CircuitByCirq

    qiskit_bc = CircuitByQiskit(expectation_calc_method="statevector")
    qulacs_bc = CircuitByQulacs()

    # cirq_bc = CircuitByCirq()
    # projectq_bc = CircuitByProjectq()   # Bugs need to fix
    # ts = CircuitByTensor()
    # qt = CircuitByQton()  #expectation_calc_method="tensor"

    ghz = GHZ_Generate(node_num=node_num,
                       p=p,
                       optimizer=optc,
                       backend=qiskit_bc)  #qt, , cirq_bc, projectq_bc, qulacs_bc
    # g = ghz.get_graph()
    # ghz._g = g
    # ghz._qc = Qcover(g, p,
    #                  optimizer=optc,
    #                  backend=qiskit_bc,
    #                  research_obj="GHZ")
    sol = ghz.run()

    print("solution is:", sol)
    params = sol["Optimal parameter value"]
    # params = [2.40756694, 2.37257408, 2.34787417, 0.79040416, 0.78994118,
    #    0.73654257]
    # ghz._qc.backend._pargs = params

    res_exp = ghz._qc.backend.expectation_calculation()
    print("optimal parameter value:", ghz._qc.backend._pargs)
    print("the optimal expectation is: ", res_exp)
    out_count = ghz._qc.backend.get_result_counts(params, ghz._g)
    import matplotlib.pyplot as plt
    from qiskit.visualization import plot_histogram
    plot_histogram(out_count)
    plt.show()

